/*
 *						cc_c
 *						~~~~
 *
 *	 Command line front end for the C68 Compilation system
 *
 *
 * (c) Copyright 1991-1996 David J. Walker
 *	   Email:  d.j.walker@x400.icl.co.uk
 *
 *	   The purpose of this module is to act as a front end
 *	   to the C development system:
 *		  a) CPP, C68, AS68/GWAA/QMAC and LD programs for C68 for QDOS.
 *		  b) CPP, C86, AS86 and LD86 for CPOC
 *
 *	   Permission to copy and/or distribute granted under the
 *	   following conditions:
 *
 *	   1). This notice must remain intact.
 *	   2). The author is not responsible for the consequences of use
 *		   this software, no matter how awful, even if they
 *		   arise from defects in it.
 *	   3). Altered version must not be represented as being the
 *		   original software.
 *
 *	This software is based on a program originally written by Jeremy Allison.
 *	It has since been heavily modified.
 */

#include "cc.h"

char _prog_name[] = CC;
char _version[]   = VERSION;
char _copyright[] = "Compiler driver";
char copyright[]  = "(c)1991-1996  D.J.Walker";

#ifdef QDOS
void (*_consetup)() = consetup_title;			/* Fancy titled screen */
int (*_cmdwildcard)() = cmdexpand;				/* Wild card expansion */
int (*_Open)(const char *, int, ...) = qopen;	/* Accept names with . */
#endif /* QDOS */

#undef _P_

struct FILE_LIST {
		struct FILE_LIST *	next;
		char *	name;
		} file_list = {NULL, NULL};

struct cpasses {
	char *p_name;			/* Program name to call for this pass */
	char *p_optbuf; 		/* Pointer to option buffers for this pass */
	char *p_suffix; 		/* Character output file name must end with */
};

int 	pass;							/* Current Pass */
struct cpasses *cp; 					/* Current Pass details */

enum pass_order {
	CC_PASS,  CPP_PASS,   C68_PASS,  ASM_PASS,
#ifdef QL
#ifdef QDOS
	QMAC_PASS, GWASS_PASS,
#endif /* QDOS */
#endif /* QL */
	LD_PASS, LDPRE_PASS, LDPOST_PASS,
	DEL_PASS, 
#ifdef QL
#ifdef QDOS
	COPY_PASS,
#endif /* QDOS */
#endif /* QL */
	MAX_PASS
	};
/*
 *	N.B.  This array must correspond to the pass_order above
 */
struct cpasses passes[] = {
	{ CC,     NULL, NULL },
	{ CPP,    NULL, "i" },
	{ COMPILER,  NULL, "s" },
	{ ASM,    NULL, "o" },          /* Standard (default) assembler */
#ifdef QL
#ifdef QDOS
	{ "qmac", NULL, "o" },          /* Quanta QMAC assembler */
	{ "gwass",NULL, "o" },          /* George GWILT special GWASS assembler */
#endif /* QDOS */
#endif /* QL */
	{ LD,     NULL,  0 },
	{ LD,     NULL,  0 },
	{ LD,     NULL,  0 },
	{ "rm",   NULL,  0 },
#ifdef QL
#ifndef QDOS
	{ "cp",   NULL,  0 }
#endif /* QDOS */
#endif /* QL */
};

#ifdef QL
int rllflag = 0;				/* Set if RLL need to be (potentially) linked in */
int rlllib	= 0;				/* Set if building RLL library */
#endif /* QL */

enum actions {
		ACTION_IGNORE,					/* Option is not used */
		ACTION_ADD, 					/* Pass option to specified pass */
		ACTION_ADD_FILES,				/* Add option to filelist */
		ACTION_ADD_REST,				/* Pass option remainder to pass */
		ACTION_ADD_ALL, 				/* Pass option to all passes */
		ACTION_ADD_STKMEM,				/* Stack/Memory option add to all passes */
		ACTION_ADD_MORE,				/* More data follows, allow for space */
		ACTION_ADD_OPTIONAL,			/* More data may follow */
		ACTION_CPP_LAST,				/* Stop after the pre-processor phase */
		ACTION_C68_LAST,				/* Stop after the compiler phase */
		ACTION_ASM_LAST,				/* Stop after the assembler phase */
		ACTION_GWASS,					/* Use GWASS assembler */
		ACTION_QMAC,					/* Use QMAC as assembler phase */
		ACTION_FPU, 					/* Target requires hardware FPU */
		ACTION_VERBOSE, 				/* Run in verbose mode */
		ACTION_VERBOSE_ALL, 			/* Run all components in verbose mode */
		ACTION_TMP, 					/* Set directory for temporary files */
		ACTION_TMP_ALL, 				/* ... temporary and target files */
		ACTION_TRAD,					/* Set -trad option for C68 */
		ACTION_PATH 				/* Set program search path */
#ifdef QL
		,ACTION_RLL 					/* Set for RLL library to be used */
		,ACTION_RLL_LIB 				/* Set for RLL library being built */
		,ACTION_RLL_DEF 				/* Set for RLL defines used */
		,ACTION_UCHAR					/* Set -uchar options for C68 */
		,ACTION_XA						/* Set -Xa options */
		,ACTION_XC						/* Set -Xc options */
		,ACTION_XT						/* Set -Xt options */
#endif /* QL */
		};

/*
 *	Options list for CC program
 *
 *	This is the complete list of options recognised by the CC program.
 *	For each option, the pass to which it relates is given, plus the
 *	action to be taken on encountering this option.  The table is
 *	searched sequentially, so if options might conflict they must be
 *	put in the table in such an order that this conflict is resolved.
 */
struct OPTIONS {
		char	*identifier;
		int 	pass;
		int 	action;
	} options[] = {
#ifdef QL
				{"%",       CC_PASS,     ACTION_ADD_STKMEM},
#endif /* QL */
#ifdef CPOC
				{"0",       ASM_PASS,    ACTION_ADD},
#endif /* CPOC */
#ifdef QL
				{"=",       CC_PASS,     ACTION_ADD_STKMEM},
				{"A",       ASM_PASS,    ACTION_ADD},
				{"BUFL",    LDPRE_PASS,  ACTION_ADD},
				{"BUFP",    LDPRE_PASS,  ACTION_ADD},
				{"bufl",    LDPRE_PASS,  ACTION_ADD},
				{"bufp",    LDPRE_PASS,  ACTION_ADD},
#endif /* QL */
				{"CODEMOD", C68_PASS,    ACTION_ADD},
				{"CRF",     LDPRE_PASS,  ACTION_ADD},
				{"C",       CPP_PASS,    ACTION_ADD},
				{"codemod", C68_PASS,    ACTION_ADD},
				{"crf",     LDPRE_PASS,  ACTION_ADD},
				{"c",       CC_PASS,     ACTION_ASM_LAST},
				{"DATAMOD", C68_PASS,    ACTION_ADD},
				{"DEBUG",   LDPRE_PASS,  ACTION_ADD_OPTIONAL},
#ifdef QL
				{"DRLL_LIBS",CPP_PASS,   ACTION_RLL_DEF},
#endif /* QL */
				{"D",       CPP_PASS,    ACTION_ADD_MORE},
				{"datamod", C68_PASS,    ACTION_ADD},
				{"debug",   LDPRE_PASS,  ACTION_ADD_OPTIONAL},
				{"d",       CPP_PASS,    ACTION_ADD},
				{"error",   C68_PASS,    ACTION_ADD},
				{"extens",  C68_PASS,    ACTION_ADD},
				{"extern",  C68_PASS,    ACTION_ADD},
				{"E",       CPP_PASS,    ACTION_CPP_LAST},
#ifdef QL
				{"fpu",     CC_PASS,     ACTION_FPU},
#endif /* QL */
#ifdef CPOC
				{"fp",      CC_PASS,     ACTION_FPU},
#endif /* CPOC */
				{"frame",   C68_PASS,    ACTION_ADD},
				{"f",       LD_PASS,     ACTION_ADD},
#ifdef QL
#ifdef QDOS
				{"GWASS",   CC_PASS,     ACTION_GWASS},
				{"gwass",   CC_PASS,     ACTION_GWASS},
#endif /* QDOS */
#endif /* QL */
				{"g",       C68_PASS,    ACTION_ADD},
				{"H",       CC_PASS,     ACTION_IGNORE},
				{"h",       CC_PASS,     ACTION_IGNORE},
				{"I",       CPP_PASS,    ACTION_ADD_MORE},
				{"icode",   C68_PASS,    ACTION_ADD},
				{"i",       CPP_PASS,    ACTION_ADD_MORE},
				{"L",       LDPRE_PASS,  ACTION_ADD_MORE},
				{"i",       CPP_PASS,    ACTION_ADD},
				{"lattice", C68_PASS,    ACTION_ADD},
				{"list",    C68_PASS,    ACTION_ADD},
				{"l",       LDPOST_PASS, ACTION_ADD_FILES},
				{"MM",      CPP_PASS,    ACTION_ADD},
				{"M",       CPP_PASS,    ACTION_ADD},
				{"maxerr",  C68_PASS,    ACTION_ADD},
				{"m",       LDPRE_PASS,  ACTION_ADD},
#ifdef QL
				{"N",       ASM_PASS,    ACTION_ADD},
#endif /* QL */
#ifdef OBSOLETE
				{"noformat",C68_PASS,    ACTION_ADD},
				{"noopt",   C68_PASS,    ACTION_ADD},
				{"nopeep",  C68_PASS,    ACTION_ADD},
				{"noreg",   C68_PASS,    ACTION_ADD},
				{"nostdinc",CPP_PASS,    ACTION_ADD},
				{"nostackopt",C68_PASS,  ACTION_IGNORE},
				{"nowarn",  C68_PASS,    ACTION_ADD},
#endif /* OBSOLETE */
				{"O",       C68_PASS,    ACTION_ADD},
				{"o",       LDPRE_PASS,  ACTION_ADD_MORE},
				{"P",       CPP_PASS,    ACTION_ADD},
				{"pedantic",CPP_PASS,    ACTION_ADD},
				{"peep",    C68_PASS,    ACTION_ADD},
				{"p",       CC_PASS,     ACTION_CPP_LAST},
				{"opt",     C68_PASS,    ACTION_ADD},
#ifdef QL
#ifdef QDOS
				{"QMAC",    CC_PASS,     ACTION_QMAC},
				{"qmac",    CC_PASS,     ACTION_QMAC},
#endif /* QDOS */
#endif /* QL */
				{"Qtrad",   C68_PASS,    ACTION_TRAD},
#ifdef QL
				{"Quchar",  C68_PASS,    ACTION_UCHAR},
#endif /* QL */
				{"Q",       C68_PASS,    ACTION_ADD_REST},
				{"reg",     C68_PASS,    ACTION_ADD},
#ifdef QL
				{"R",       LDPRE_PASS,  ACTION_RLL_LIB},
				{"r",       CC_PASS,     ACTION_RLL},
#endif /* QL */
				{"SYM",     LDPRE_PASS,  ACTION_ADD},
				{"S",       CC_PASS,     ACTION_C68_LAST},
				{"short",   C68_PASS,    ACTION_ADD},
				{"small",   C68_PASS,    ACTION_ADD},
				{"stackcheck",C68_PASS,  ACTION_ADD},
				{"stackopt",C68_PASS,    ACTION_ADD},
				{"sym",     LDPRE_PASS,  ACTION_ADD},
				{"s",       LDPRE_PASS,  ACTION_ADD_MORE},
				{"TMP",     CC_PASS,     ACTION_TMP_ALL},
				{"TRAD",    C68_PASS,    ACTION_TRAD},
				{"tmp",     CC_PASS,     ACTION_TMP},
				{"trace",   C68_PASS,    ACTION_ADD},
				{"trad",    C68_PASS,    ACTION_TRAD},
				{"trigraphs",CPP_PASS,   ACTION_ADD},
#ifdef OBSOLETE
				{"UNPROTO", CC_PASS,     ACTION_IGNORE},
#endif /* OBSOLETE */
				{"U",       CPP_PASS,    ACTION_ADD},
#ifdef QL
				{"uchar",   C68_PASS,    ACTION_UCHAR},
#endif /* QL */
				{"undef",   CPP_PASS,    ACTION_ADD},
#ifdef OBSOLETE
				{"unproto", CC_PASS,     ACTION_IGNORE},
#endif /* OBSOLETE */
				{"u",       CC_PASS,     ACTION_IGNORE},
				{"U",       CPP_PASS,    ACTION_ADD_MORE},
				{"v",       CC_PASS,     ACTION_VERBOSE_ALL},
				{"V",       CC_PASS,     ACTION_VERBOSE},
				{"warn",    C68_PASS,    ACTION_ADD},
#ifdef QL
				{"Xa",      CC_PASS,     ACTION_XA},
				{"Xc",      CC_PASS,     ACTION_XC},
				{"Xt",      CC_PASS,     ACTION_XT},
#endif /* QL */
				{"Y",       CC_PASS,     ACTION_PATH},
				{"y",       CC_PASS,     ACTION_PATH},
				{NULL, 0, 0 }
				};

char * tmp_dir = "";

int isname = 0; 		   /* set as soon as filename found */
int last_pass = ASM_PASS;
int asmpass = ASM_PASS;    /* Set to identify the assembler to be used */
int dolink = 1;

int  verbose = 0;
int  alltemp = 0;
int  no_c68  = 0;			/* set to skip C68 pass */
#define ASMCPP_OPT	 "-P"	/* Option to CPP for assembler files */
int  fpflag  = 0;			/* Floating Point support required */

char *ccpath	= NULL; 	/* Path for finding CC phases */
char *ccinc 	= NULL; 	/* Path for automatic include path parameter */
char *cclib 	= NULL; 	/* Path for automatic library path parameter */



/*============================================================ WHAT_OPT */
struct OPTIONS	*what_opt ( opt )
/*				 ~~~~~~~~
 *	Routine to decide if an option is allowable.
 *
 *	Returns NULL on failure, and pointer into options array otherwise.
 *----------------------------------------------------------------------*/
char *opt;
{
	struct OPTIONS *ptr;
	int 	i;

	DBG(("WHAT_OPT",0x18,"Enter: opt='%s'",opt));
	for( i=0, ptr= &options[0] ; ptr->identifier != NULL; ptr++, i++ ) {
		DBG(("WHAT_OPT",0x18,"comparing '%s' with '%s'",opt, ptr->identifier));
		if( !strncmp( opt, ptr->identifier, strlen(ptr->identifier))) {
			DBG(("WHAT_OPT",0x18,"Exit %p",ptr));
			return (ptr);
		}
	}
	DBG(("WHAT_OPT",0x18,"Exit NULL: Option not recognised"));
	return (NULL);
}


#ifdef QDOS
/*============================================================== NO_CONTINUE */
int 	no_continue ()
/*		~~~~~~~~~~~
 *
 * Routine to check if this compilation set should be aborted.
 *---------------------------------------------------------------------------*/
{
	extern long _endchanid;
	chanid_t chan = getchid(STDIN_FILENO);
	int  errnosav = errno;
	int  oserrsav = _oserr;
	char c;
	int err;

	if (! _endchanid) 
		return 1;
	printstr ("Compilation failed\nPress <Enter> to exit, any key to continue\n", NULL);
	(void) sd_cure( chan, (timeout_t)(-1)); /* Ensure a cursor is enabled */
	if( (err = io_fbyte( chan, (timeout_t)(-1), &c)) != 0 ) {
		errno = errnosav;
		exit( err );
	}
	errno = errnosav;
	_oserr = oserrsav;
	return ( c == '\n');
}
#endif /* QDOS */

/*================================================================= ADD_TMP */
void	add_tmp (char *outname, char *suffix, char *last_suffix)
/*		~~~~~~~
 *
 *	Routine to compose an output filename with the right extension. 
 *	Does not prepend tmp_dir if suffix is equal to last suffix 
 *	desired.
 *	DJW 07/06/91	unless 'alltemp' set.
 *---------------------------------------------------------------------------*/
{
	char tmpname[MAX_FNAME];
#ifdef QDOS
	int extra;
#endif /* QDOS */

	DBG(("ADD_TMP",0x12,"enter: tmp_dir='%s', outname='%s', suffix='%s', last_suffix='%s'",tmp_dir,outname, suffix, last_suffix));
	(void)strcpy (strrpbrk(outname,"._")+1,suffix);
	DBG(("ADD_TMP",0x18,"new outname = '%s'", outname));
	if (alltemp || strcmp(suffix,last_suffix) ) {
		DBG(("ADD_TMP",0x14,"setting temp directory to '%s'",tmp_dir));
		(void) strcpy( tmpname, tmp_dir);
#ifdef QDOS
		if (isdevice(outname, &extra) && strchr(outname,'_')) {
			DBG(("ADD_TMP",0x14,"removing device name from '%s'",outname));
			(void) strcat (tmpname, (char *)((long)strchr(outname, '_')+1));
		} else {
			(void) strcat( tmpname, outname);
		}
		DBG(("ADD_TMP",0x14," ...giving '%s'",tmpname));
	} else {
		(void) _mkname (tmpname, outname);
#else
		{
			char *p = (char *) FilePart(outname);
			AddPart(tmpname, p, MAX_FNAME);
		}
		DBG(("ADD_TMP",0x14," ...giving '%s'",tmpname));
	} else {
		(void) strcpy (tmpname, outname);
#endif /* QDOS */
	}
	(void) strcpy( outname, tmpname);
	DBG(("ADD_TMP",0x12,"exit: outname($%p)='%s'",outname,outname));
	return;
}

/*================================================================= ADD_OPT */
void	add_option ( pass_opt, p)
/*		~~~~~~~~~~
 *
 *	Add an option to an option buffer.
 *	The buffer will be expanded to make enough room.
 *
 *	Except for the GWASS assembler, a space is
 *	always added before a new option.
 *-------------------------------------------------------------------------*/
int 	pass_opt;
char	*p;
{
	char	**ptr;
	size_t	oldsize;			/* Size of current command line */
	size_t	newbit; 			/* Size of string to be added */
	size_t	spacesize;			/* Count of spaces (if any) as separators */

	DBG(("ADD_OPT",0x14,"pass = %d; p='%s'",pass_opt, p));

	ptr = &passes[pass_opt].p_optbuf;
	DBG(("ADD_OPT",0x18,"   current option buffer='%s'",*ptr));
	DBG(("ADD_OPT",0x18,"   new option to add='%s'",p));
	oldsize = strlen(*ptr);
	newbit = strlen(p);
#ifdef QL
	spacesize = (oldsize == 0 || pass_opt == GWASS_PASS) ? 0 : 1;
#else
	spacesize = (oldsize == 0) ? 0 : 1;
#endif /* QDOS */

	if ((*ptr = realloc (*ptr, (size_t)(oldsize + newbit + spacesize + 1))) == NULL) {
		printstr("Failure extending option buffers\n", NULL);
		exit (ERR_OM);
	}

	if (spacesize) {
		(void) strcat( *ptr, " ");
	}
	(void) strcat( *ptr, p );
	DBG(("ADD_OPT",0x14,"Exit:  new option buffer='%s'",*ptr));
	return;
}


/*=============================================================== BASENAME */
char   *basename ( fname )
/*		~~~~~~~~
 *
 * Routine to return the base name of any filename. 
 * (Cheats in this implementation as we know we should
 * end in "_ext".  We actually also allow for ".ext".
 *------------------------------------------------------------------------*/
char *fname;
{
#ifndef QDOS
	return FilePart(fname);
#else
	int got_under;

	char *p = &fname[strlen(fname)-1];

	for( got_under = 0; p > fname && got_under < 2; p--) {
		if( strchr("_.",*p)) {
			got_under++;
		}
	}
	return (p == fname ? p : p+2);
#endif /* QDOS */
}


/*=========================================================== COMMAND_LINE */
int 	command_line (p_name, optbuf, inname, outname)
/*		~~~~~~~~~~~~
 *
 *	Used to print a command line if verbose specified.	The
 *	strings supplied or printed space seperated.
 *
 *	It will also add any automatic parameters to command lines.
 *-------------------------------------------------------------------------*/
char	*p_name;
char	*optbuf;
char	*inname;
char	*outname;
{
	char *	p;
	int 	ret;
	char *	cmdbuf;
	size_t	cmdlen;

	DBG(("COMMAND_LINE",0x14,"p_name='%s' (pass=%d)",p_name,pass));
	DBG(("COMMAND_LINE",0x14,"optbuf/preld   ='%s'",optbuf));
	DBG(("COMMAND_LINE",0x14,"inname/linkfile='%s'",inname));
	DBG(("COMMAND_LINE",0x14,"outname/postld ='%s'",outname));

	/*
	 *	Initialise with length of program
	 *	name plus a small overhead.
	 */
#ifdef QDOS
	cmdlen = strlen(p_name) + 10;
#else
	cmdlen = strlen(ccpath) + strlen(p_name) + 10;
#endif /* QDOS */
	/*
	 *	Now for the command line options.
	 *	Allow for fact that certain passes
	 *	add in some automatic parameters.
	 */
	switch (pass)
	{
		case CPP_PASS:
				cmdlen += strlen(ccinc) + 3;
				break;
		case C68_PASS:
				if (fpflag)
				{
					cmdlen += sizeof("-fpu=yes ");
				}
				break;
		case LD_PASS:
				cmdlen += strlen(cclib) + 4;
#ifdef QL
				if (fpflag)
				{
					cmdlen += sizeof(" -lfpu");
				}
				cmdlen += sizeof(" -lc");
#endif /* QL */
#ifdef CPOC
				cmdlen += fpflag ? sizeof (" fpinit.o") : sizeof(" nofpinit.o");
				cmdlen += sizeof (" -lp");
#endif /* CPOC */
				break;
	}
	cmdlen += strlen(optbuf);
	/*
	 *	Calculate the command line space for the filenames
	 */
	switch (pass)
	{
#ifdef QL
#ifdef QDOS
		case GWASS_PASS:
			/*
			 *	For GWASS the input name must include the full path,
			 *	and the output name is not passed in the command line.
			 */
			cmdlen += 40 - strlen(inname);
			break;
#endif /* QDOS */
#endif /* QL */
		default:
			cmdlen += strlen(inname) + strlen(outname);
			break;
	}

	DBG(("COMMAND_LINE",0x14," ...allocating %d for command line buffer", cmdlen));
	p = cmdbuf = malloc (cmdlen);

	/*
	 *	Set up program name
	 *	For non-QDOS systems include the path (if any)
	 */
#ifndef QDOS
	if (strlen(ccpath))
	{
		p = stpcpy(p, ccpath);
		if (*(p-1) != DSEP)
		{
			*p++ = DSEP;
		}
	}
#endif /* ! QDOS */
	p = stpcpy(p, p_name);
	*p++ = ' ';

	/*
	 *	Copy across options
	 *	First add any 'automatic' parameters
	 *	that are dependant on the pass.
	 */
	switch (pass)
	{
		case CPP_PASS:
				if (*ccinc)
				{
					p = stpcpy(p,"-I");
					p = stpcpy(p,ccinc);
					*p++ = ' ';
					DBG(("COMMAND_LINE",0x14," ...adding INC option to give '%s'", cmdbuf));
				}
				break;
		case C68_PASS:
				if (fpflag)
				{
					p = stpcpy(p,"fpu=yes ");
				}
				break;
		case LD_PASS:
				if (*cclib)
				{
					p = stpcpy(p,"-L");
#ifdef CPOC
					*p++ = ' ';
#endif /* CPOC */
					p = stpcpy(p,cclib);
					*p++ = ' ';
					DBG(("COMMAND_LINE",0x14," ...adding LIB option to give '%s'", cmdbuf));
				}
				break;
	}
	p = stpcpy (p, optbuf);

	DBG(("COMMAND_LINE",0x14," ...added optbuf parameter to give '%s'", cmdbuf));

	/*
	 *	Add input filename.
	 */
	if (inname && *inname)
	{
		switch (pass)
		{
#ifdef QL
#ifdef QDOS
			/*
			 *	GWASS has only an input name,
			 *	it must include the full path,
			 *	and no space seperator from options.
			 */
			case GWASS_PASS:
				(void) _mkname (p, inname);
				p += strlen(p);
				break;
#endif /* QDOS */
#endif /* QL */
			default:
				*p++ = ' ';
				p = stpcpy (p, inname);
				break;
		}
	}
	DBG(("COMMAND_LINE",0x14," ...added input filename to give '%s'", cmdbuf));

	/*
	 *	Add output filename.
	 */
	if (outname && *outname)
	{
		switch (pass)
		{
#ifdef QL
#ifdef QDOS
			/*
			 *	GWASS does not have one.
			 */
			case GWASS_PASS:
				break;
#endif /* QDOS */
#endif /* QL */
			/*
			 *	Certain passes need a -o preceeding the name.
			 */
#ifndef QL
#ifdef __unix__
			case CPP_PASS:
#endif /* __unix__ */
			case ASM_PASS:
				*p++ = ' ';
				*p++ = '-';
				*p++ = 'o';
				/* FALLTHRU */
#endif /* QL */
			default:
				*p++ = ' ';
				p = stpcpy (p,outname);
				break;
		}
	}
	/*
	 *	Add any autoamtic libraries
	 */
	switch (pass)
	{
		case LD_PASS:
#if QL
			if (fpflag)
			{
				p = stpcpy(p, " -lfpu");
			}
			p = stpcpy(p, " -lc");

#endif /* QL */
#if CPOC
			if (fpflag)
			{
				p = stpcpy(p, " fpinit.o");
			}
			else
			{
				p = stpcpy(p, " nofpinit.o");
			}
			p = stpcpy(p," -lp");
#endif /* CPOC */
			break;
	}
	*p = '\0';		/* NULL terminate */
	DBG(("COMMAND_LINE",0x14," ...added output filename to give '%s'", cmdbuf));

	/*
	 *	Print details of what we are about to run.
	 */
	if (verbose)
	{
		(void) printstr("  ",cmdbuf,"\n", NULL);
	}

	/*
	 *	Now run the command we have built up.
	 */
#ifdef QL
#ifdef QDOS
	if (pass == GWASS_PASS)
	{
		/*
		 *	With GWASS on QDOS we use the execl() call
		 *	as we need to pass exactly one channel to
		 *	be used for output messages.
		 */
		int 	fd[2];
		fd[0] = 1;						/* Number of channels to pass */
		fd[1] = STDERR_FILENO;			/* Error channel */
		p = strchr(cmdbuf,' ');
		*p++ = '\0';
		ret = execl(cmdbuf, fd, p_name, p, NULL);
	}
	else
#endif /* QDOS*/
#endif /* QL */
		ret = system(cmdbuf);

	free (cmdbuf);
	DBG(("COMMAND_LINE",0x11,"Exit: ret=%d",ret));
	return (ret);
}


/*================================================================== USEAGE */
void	useage()
/*		~~~~~~
 *	Simply print a usage message.
/*--------------------------------------------------------------------------*/
{
	(void) printstr ("useage: ",CC," [options] filename[s]\n",NULL);
	exit( -1 );
}


/*============================================================== PARAMETERS */
void	Parameters (int argc, char * argv[])
/*		~~~~~~~~~~
 *	Process a set of parameters.  These may have been passed via the
 *	command line, or via an environment variable - we do not care.
 *--------------------------------------------------------------------------*/
{
	char workbuf[100];
	char *argvsf;
	int sf;

	DBG(("PARAMETERS",0x11,"Enter: argc=%d", argc));
	for( sf = 1; sf < argc; sf++) {
		struct OPTIONS *optptr;

		DBG(("PARAMETERS",0x12,"parameter parsing 'sf' loop - argv[sf]='%s'", \
										argv[sf]));
		if( *argv[sf] != '-' ) {
			struct FILE_LIST *ptr;
			DBG(("PARAMETERS",0x12,"filename - adding to list"));
			isname++;
ADD_FILELIST:
			for (ptr = &file_list ; ptr->next ; ptr=ptr->next)
				;
			DBG(("PARAMETERS",0x24,"file_list starts at %p, finishes at %p", \
				  &file_list, ptr));
			ptr->next = malloc (sizeof(struct FILE_LIST));
			DBG (("PARAMETERS",0x24,"Added new FILE_LIST structure at address %p", \
											ptr->next));
			if ((ptr=ptr->next) == NULL)
			{
				DBG(("PARAMETERS",0x21,"Failed to allocate FILE_LIST structure"));
				exit (ERR_OM);
			}
			ptr->next = NULL;
			ptr->name = argv[sf];
			continue; /* Skip filenames */
		}

		/* Stop after getting options */
		argvsf = argv[sf];

		/*
		 *	Identify the option.
		 *	The value of returned as an index into
		 *	the option array.
		 */
		DBG (("PARAMETERS",0x28,"last_pass = %d",last_pass));
		if ((optptr = what_opt( &argvsf[1] )) == NULL ) {
			DBG(("PARAMETERS",0x12,"   option not recognised - so ignored"));
			printstr ("option '",argvsf,"' not recognised - ignored\n",NULL);
			continue;
		}
		DBG(("PARAMETERS",0x12," action=%d, pass=%d",optptr->action,optptr->pass));
		switch (optptr->action) {
#ifdef QL
		case ACTION_ADD_STKMEM:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_STKMEM)"));
				argvsf++;		/* Skip over - character */
				add_option (CPP_PASS, argvsf);
				add_option (C68_PASS, argvsf);
				add_option (ASM_PASS,argvsf);
				add_option (LD_PASS,  argvsf);
				break;				
#endif /* QL */

		case ACTION_ADD_ALL:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_ALL)"));
				add_option (CPP_PASS, argvsf);
				add_option (C68_PASS, argvsf);
				add_option (ASM_PASS,argvsf);
				add_option (LD_PASS,  argvsf);
				break;				

		case ACTION_ADD:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD)"));
				add_option (optptr->pass,argvsf);
				break;

		case ACTION_ADD_FILES:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_FILEIST)"));
				goto ADD_FILELIST;

		case ACTION_ADD_REST:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_REST)"));
				(void) strcpy(workbuf,"-");
				add_option (optptr->pass, strcat(workbuf, argvsf+2));
				break;

		case ACTION_ADD_MORE:
#ifdef QL
ADD_MORE:
#endif /* QL */
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_MORE)"));
				/*
				 *	There must be more (typically a filename).
				 *	If necessary get the next parameter
				 */
				if (argvsf[2] == '\0')
				{
					(void) strcpy (workbuf,argvsf);
					argvsf = argv[++sf];
#ifdef CPOC
					(void) strcat(workbuf, " ");
#endif /* CPOC */
					(void) strcat(workbuf, argvsf);
					add_option (optptr->pass, workbuf);
					break;
				}
				add_option (optptr->pass,argvsf);
				break;

		case ACTION_ADD_OPTIONAL:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_OPTIONAL)"));
				add_option (optptr->pass,argvsf);
				if (*(argv[sf+1]) != '-') {
					add_option (optptr->pass, argv[++sf]);
				}
				break;

		case ACTION_IGNORE:
				DBG(("PARAMETERS",0x12,"   (ACTION_IGNORE)"));
				break;

		case ACTION_TMP_ALL:
				DBG(("PARAMETERS",0x12,"   (ACTION_TMP_ALL)"));
				alltemp = 1;
				/* FALLTHRU */
		case ACTION_TMP: /* Define temp directory */
				DBG(("PARAMETERS",0x12,"   (ACTION_TMP)"));
				/* Ensure directory ends in '_' */
				if( argvsf[strlen(argvsf)-1] != '_' )
#ifdef QDOS
				{
					printstr("Temporary directory (", argvsf, ") must end in _\n", NULL);
					exit (ERR_BP);
				}
#endif /* QDOS */
				tmp_dir = &argvsf[4];
				DBG(("PARAMETERS",0x12,"   tmp_dir=%s",tmp_dir));
				break;

		case ACTION_VERBOSE_ALL:
				DBG(("PARAMETERS",0x12,"   (ACTION_VERBOSE_ALL)"));
				verbose++;
#ifndef CPOC
				add_option (CPP_PASS, argvsf);
#endif /* CPOC */
				add_option (C68_PASS, argvsf);
#ifndef CPOC
				add_option (ASM_PASS,argvsf);
				add_option (LDPRE_PASS,argvsf);
#endif /* CPOC */
				break;				

		case ACTION_VERBOSE:
				DBG(("PARAMETERS",0x12,"   (ACTION_VERBOSE)"));
				verbose++;
				break;

		case ACTION_CPP_LAST:
				DBG(("PARAMETERS",0x12,"   (ACTION_CPP_LAST)"));
				add_option (CPP_PASS, argvsf);
				last_pass = CPP_PASS;
				dolink = 0;
				break;

		case ACTION_C68_LAST:
				DBG(("PARAMETERS",0x12,"   (ACTION_C68_LAST)"));
				last_pass = C68_PASS;
				dolink = 0;
				break;

		case ACTION_ASM_LAST:
				DBG(("PARAMETERS",0x12,"   (ACTION_ASM_LAST)"));
				last_pass = ASM_PASS;
				dolink = 0;
				break;

		case ACTION_TRAD:
				DBG(("PARAMETERS",0x12,"   (ACTION_TRAD)"));
				add_option( CPP_PASS, "-U__STDC__");
				add_option (optptr->pass,argvsf);
				break;

		case ACTION_FPU:
				DBG(("PARAMETERS",0x12,"   (ACTION_FPU)"));
				fpflag++;
#ifdef CPOC
				break;
#endif /* CPOC */
#ifdef QL
#ifdef QDOS
				if (CheckCPU() < 0x20) {
					printstr ("The '-fpu' option needs the GWASS assembler\n", NULL);
					printstr ("The GWASS assembler requires a 68020 or better\n", NULL);
					exit (-1);
				}
				/* FALLTHRU */

		case ACTION_GWASS:
				DBG(("PARAMETERS",0x12,"   (ACTION_GWASS)"));
				if (CheckCPU() < 0x20) {
					printstr ("The GWASS assembler requires a 68020 or better\n", NULL);
					exit(-1);
				}
				add_option (C68_PASS,"-qmc68k");
				add_option (GWASS_PASS,"D/5/Q//A/1/");
				asmpass = GWASS_PASS;
				passes[C68_PASS].p_suffix = "asm";
				break;

		case ACTION_QMAC:
				DBG(("PARAMETERS",0x12,"   (ACTION_QMAC)"));
				asmpass = QMAC_PASS;
				passes[C68_PASS].p_suffix = "asm";
				break;
#endif /* QDOS */
#endif /* QL */

		case ACTION_PATH:
				DBG(("PARAMETERS",0x12,"   (ACTION_PATH)"));
				ccpath = &argvsf[2];
				break;
#ifdef QL
		case ACTION_RLL_DEF:
				DBG(("PARAMETERS",0x12,"    (ACTION_RLL_DEF)"));
				rllflag++;
				goto ADD_MORE;

		case ACTION_RLL_LIB:
				DBG(("PARAMETERS",0x12,"    (ACTION_RLL_LIB)"));
				rlllib++;
				goto ADD_MORE;

		case ACTION_RLL:
				DBG(("PARAMETERS",0x12,"    (ACTION_RLL)"));
				add_option (CPP_PASS, "-DRLL_LIBS");
				rllflag++;
				goto ADD_FILELIST;

		case ACTION_UCHAR:
				DBG(("PARAMETERS",0x12,"   (ACTION_UCHAR)"));
				add_option (CPP_PASS,"-D__CHAR_UNSIGNED__");
				add_option (optptr->pass,argvsf);
				break;
		case ACTION_XA:
				add_option (LDPRE_PASS, "-lxa");
				add_option (CPP_PASS, "-D__STDC__=0");
				add_option (CPP_PASS, "-trigraphs");
				break;
		case ACTION_XC:
				add_option (C68_PASS, "-obsolete");
				add_option (CPP_PASS, "-pedantic");
				add_option (CPP_PASS, "-D__STRICT_ANSI__= ");
				add_option (LDPRE_PASS, "-lxc");
				add_option (CPP_PASS, "-D__STDC__=1");
				add_option (CPP_PASS, "-trigraphs");
				break;
		case ACTION_XT:
				add_option (LDPRE_PASS, "-lxt");
				add_option (CPP_PASS, "-D__STDC__=0");
				break;
#endif /* QL */
		default:
				DBG(("PARAMETERS",0x12,"   (default)"));
				useage();
				break;
		} /* end of 'index' switch */
		/*
		 *	Option has been eaten - mark space unused
		 */
		argv[sf] = NULL;
	}	/* end of 'options' for loop */
	return;
}


#ifdef ENVVAR
/*================================================== ENVIRONMENT_PARAMETERS */
static
void	Environment_Parameters (char * argv[])
/*		~~~~~~~~~~~~~~~~~~~~~~
 *	This routine is responsible for getting the CC_OPTS environment
 *	variable, and then using it to set up data structures that look
 *	like the 'argc' and 'argv' ones passed to main(), before calling
 *	the shared parameter analysis routine.
 *-------------------------------------------------------------------------*/
{
	char	*cc_opts;
	int 	envc;
	char ** envv;

#ifdef QL
	cc_opts = getenv("CC_OPTS");
#endif /* QL */
#ifdef CPOC
	cc_opts = getenv("CPOC_OPTS");
#endif /* CPOC */
	if (cc_opts == NULL || *cc_opts == '\0')
	{
		return;
	}
	/*
	 *	Initialise vector and count
	 */
	envc = 1;
	envv = (char **)malloc (2 * sizeof(char *));
	envv[0] = argv[0];
	envv[1] = NULL;

	/*
	 *	Now parse the Environment Variable, splitting it
	 *	up into NULL terminated strings and building the
	 *	vector array.
	 */
	for ( ;  ; ) {
		char sc;
		/*
		 *	If we have any leading whitespace characters
		 *	then we need to skip over them.
		 */
		while (*cc_opts == ' ') {
			cc_opts++;
		}
		if (*cc_opts == '\0') {
			break;
		}
		sc = ' ';
		/*
		 *	Look for any of the characters which are
		 *	stored, but treated specially.
		 */
		switch (*cc_opts) {
		case '\'' :
		case '"'  :
				/* Next argument is string - store terminator */
				sc = *cc_opts++;
				break;
		case '\\' :
				cc_opts++;
				break;
		}
		/*
		 *	There is at least one more element, so add this
		 *	new element to the vector array.
		 */
		envv = (char **)realloc (envv, (size_t)((envc + 2) * sizeof (char *)));
		envv[envc++] = cc_opts;
		envv[envc] = NULL;
		/*
		 * Parse until terminating character reached
		 * The terminating character is changed to a NULL byte.
		 * Allow for premature termination by NULL byte.
		 */
		for ( ;  ; cc_opts++ ) {
			if (*cc_opts == '\0') {
				break;
			}
			if (*cc_opts == sc) {
				*cc_opts++ = '\0';
				break;
			}
		}
	} /* End of main for loop */

	Parameters (envc,envv);
	return;
}


/*==================================================== ENVIRONMENT_VARIABLES */
void	Environment_Variables (void)
/*		~~~~~~~~~~~~~~~~~~~~~
 *	This routine is used to read various environment variables and set a
 *	global variable appropriately depending on whether the environment
 *	variable was set or not.
 *
 *	N.B.  It is important that the value of an environment variable is
 *		  copied to a new string as changing any environment variable
 *		  can cause the others to be moved.
 *--------------------------------------------------------------------------*/
{
	static struct {
		char *	name;
		char ** result;
		char *	deflt;
	} *ptr, envtable[] = {
		{"TMP",     &tmp_dir, ""},
		{"TEMP",    &tmp_dir, ""},
#ifdef QL
		{"CC_PATH", &ccpath, ""},
		{"CC_CPP",  &passes[CPP_PASS].p_name, CPP},
		{"CC_C68",  &passes[C68_PASS].p_name, "c68"},
		{"CC_ASM",  &passes[ASM_PASS].p_name,"as68"},
		{"CC_LD",   &passes[LD_PASS].p_name,  LD},
		{"CC_INC",  &ccinc, NULL},
		{"CC_LIB",  &cclib, NULL},
#endif /* QL */
#ifdef CPOC
		{"CPOC_PATH", &ccpath, ""},
		{"CPOC_CPP",  &passes[CPP_PASS].p_name, CPP},
		{"CPOC_C86",  &passes[C68_PASS].p_name, "c86"},
		{"CPOC_ASM",  &passes[ASM_PASS].p_name,"as86"},
		{"CPOC_LD",   &passes[LD_PASS].p_name,  LD},
		{"CPOC_INC",  &ccinc, ""},
		{"CPOC_LIB",  &cclib, ""},
#endif /* CPOC */
		{NULL, NULL, NULL}};

	char * result;

	DBG(("ENVIRONMENT_VARIABLES",0x21,"Enter"));
	for (ptr = &envtable[0] ; ptr->name != NULL ; ptr++)
	{
		DBG(("ENVIRONMENT_VARIABLES",0x24,"looking for %s", ptr->name));
		if ((result = getenv(ptr->name)) !=NULL && strlen(result))
		{
			DBG(("ENVIRONMENT_VARIABLES",0x24,"...found, value='%s' stored at 0x%p", \
										result, ptr->result));
			*(ptr->result) = strdup(result);
		}
		else
		{
			DBG(("ENVIRONMENT_VARIABLES",0x24,"... not found, default='%s' stroed at 0x%p", \
										ptr->deflt, ptr->result));
		*(ptr->result) = ptr->deflt;
		}
	}
	DBG(("ENVIRONMENT_VARIABLES",0x21,"Exit"));
	return;
}
#endif /* ENVVAR */

 
/*==================================================================== MAIN */
void	main ( argc, argv)
/*		~~~~
 *
 * Main routine to do the compilations, managing all arguments. 										  *
 *--------------------------------------------------------------------------*/
int argc;
char *argv[];
{
	struct FILE_LIST *nameptr;
	char inname[MAX_FNAME], outname[MAX_FNAME];
	char *p, *p1;
	int 	n;
	int 	ret;

	DBG_INIT();
#ifdef QDOS
	if (iscon(getchid(1),0) && !isnoclose(1))
	{
		verbose++;
	}
#endif /* QDOS */
	/*
	 *	Initialise options buffers
	 */
	for (n = 0 ; n < MAX_PASS ; n++)
	{
		if ((p = malloc((size_t)1)) == NULL)
		{
			exit (ERR_OM);
		}
		*p = '\0';
		passes[n].p_optbuf = p;
	}
#ifdef CPOC
	/*
	 *	Set up mandatory parameters
	 */
	add_option (C68_PASS,  "-int=16 -pointer=16 -separate=yes");
	add_option (ASM_PASS,  "-0 -j");
	add_option (LDPRE_PASS,"-0 -i");
#endif /* CPOC */

#ifdef ENVVAR
	Environment_Variables();
	Environment_Parameters(argv);
#endif /* ENVVAR */
	Parameters (argc, argv);
	DBG (("MAIN",0x21,"Finished parsing parameters"));

#ifdef QDOS
	if (verbose) {
		DBG(("MAIN",0x14,"CC running in VERBOSE mode"));
		printstr (" ", _prog_name, _version, " (",_copyright, ") ", copyright, "\n", NULL);
	}
#endif /* QDOS */
	/*
	 *	Now do some consistency checks
	 */
	DBG(("MAIN",0x18,"Checking that isname set (=%d)",isname));
	if( isname == 0 ) {
		DBG(("MAIN",0x14,"No filenames supplied to compile"));
		printstr("No filenames to compile\n", NULL);
		exit (ERR_BP);
	}

#ifdef QDOS
	DBG(("MAIN",0x18,"Checking tmp_dir - %p (%s)", tmp_dir, tmp_dir));
	if ( (n=strlen(tmp_dir)) != 0 ) {
		DBG(("MAIN",0x18,"tmp_dir set - checking ends in underscore"));
		if (tmp_dir[n-1] != '_' ) {
			printstr("Directory for temporary files (",tmp_dir,") must end in _\n",NULL);
			exit (ERR_BP);
		}
	}
#endif /* QDOS */

	DBG(("MAIN",0x18,"Checking ccpath - %p (%s)", ccpath, ccpath));
	if (strlen(ccpath)) {
		DBG(("MAIN",0x18,"ccpath set - checking ends in underscore"));
		if (ccpath[strlen(ccpath)-1] != DSEP ) {
			DBG(("MAIN",0x18,"... failed (%s)",ccpath));
			printstr("Directory for phases (",ccpath,") must end in ",DSEPS,"\n", NULL);
			exit (ERR_BP);
		}
#ifdef QDOS
		DBG(("MAIN",0x18,"ccpath set - need to set as program directory"));
		if (chpdir (ccpath)) {
			DBG(("MAIN",0x18,"... failed"));
			printstr("Failed to set Directory Path for C68 phases to ",ccpath,"\n",NULL);
			exit (ERR_BP);
		}
#endif /* QDOS */
	}

	if (last_pass == ASM_PASS) {
		DBG(("MAIN",0x18,"assembler is the last pass"));
		DBG(("MAIN",0x18,"... so ensure correct assembler used"));
		last_pass = asmpass;
	}
#ifdef QL
	if (rllflag)
	{
		add_option (LDPOST_PASS,"-rc");
	}
#endif /* QL */
	/*
	 *	We have now finished handling all the parameters,
	 *	so it is time to start compiling programs
	 */
	DBG(("MAIN",0x14,"Start compilation phase"));
	DBG(("MAIN",0x18,"... first FILE_LIST structure at %p", file_list.next));
	for( nameptr = file_list.next; nameptr; nameptr = nameptr->next) {
		int 	start_pass;
		DBG(("MAIN",0x14,"Next file to process is '%s'",nameptr->name));
		if (nameptr->name[0] == '-')
		{
			DBG(("MAIN",0x14,"... library file, so simply add to list link files '%s'"));
			add_option( LD_PASS, nameptr->name);
			continue;
		}
		(void) strcpy( inname, nameptr->name ); /* set initial input filename */
#ifdef QDOS
		p1 = &inname[strlen(inname)-2]; /* Point at the end of the input name */
		if (*p1 == '.') {               /* Is seperator a fullstop ? */
			*p1 = '_';					/* ... yes then change to underscore */
		}
		if(*p1++ != '_') {
			printstr("Filename '",inname,"' has no extension\n", NULL);
			exit (ERR_BP);
		}
#endif /* QDOS */

#ifndef QDOS
		p1 = inname + strlen(inname) - 2;
		if(*p1++ != '.') {
			printstr("Filename '",inname,"' has no extension\n", NULL);
			exit (ERR_BP);
		}
#endif /* ! QDOS */

		switch( toupper((unsigned char)*p1) ) {
		case 'C':
				DBG(("MAIN",0x18,"File is a _C file"));
				start_pass = CPP_PASS;
				break;
		case 'I':
				DBG(("MAIN",0x18,"File is a _I file"));
				start_pass = C68_PASS;
				break;
		case 'S':
				DBG(("MAIN",0x18,"File is a _S file"));
ASM_PASS:
				start_pass = asmpass;

				if (asmpass == ASM_PASS) {
					int  fd;
					char ch;

					DBG(("MAIN",0x18,"...using AS68/AS86 assembler"));
					/*
					 *	This is to allow for the special case of an
					 *	assembler file that starts with a # character
					 *	as this means it needs to be pre-processed
					 */
					if ((fd = open(inname,O_RDONLY)) > 0) {
						(void) read (fd, &ch, 1);
						(void) close (fd);
						if (ch == '#') {
							no_c68 = 1;
							start_pass = CPP_PASS;
							add_option (CPP_PASS, ASMCPP_OPT);
						}
					}
				}
				break;
		case 'X':
				DBG(("MAIN",0x18,"File is a _X file"));
				no_c68 = 1;
				start_pass = CPP_PASS;
				add_option (CPP_PASS, ASMCPP_OPT);
#ifdef QL
#ifdef QDOS
				passes[CPP_PASS].p_suffix = (asmpass == ASM_PASS) ? "i" : "asm";
#else
				passes[CPP_PASS].p_suffix = "i";
#endif /* QDOS */
#endif /* QL */
				break;
		case 'O':
				DBG(("MAIN",0x18,"File is a _O file"));
				start_pass = LD_PASS;
				add_option( LD_PASS, inname);
				break;
		case 'A':
				DBG(("MAIN",0x18,"File might be an _ASM file"));
				if (!stricmp(p1,"asm")) {
					goto ASM_PASS;
				}
				/* FALL THRU */
		default:
			printstr("Unknown extension '_",*p1,"'\n",NULL);
			exit (ERR_BP);
		}
		DBG (("MAIN",0x18,"start_pass=%d, last_pass=%d",start_pass,last_pass)); 
		for(pass=start_pass ; (pass <= last_pass) && (pass < LD_PASS) ; pass++) {
			/*
			 *	This is to allow for the fact that _s files
			 *	that start with # in the first character are
			 *	put through the pre-processor first
			 */
			if ((pass == C68_PASS && no_c68)
			||	(pass == ASM_PASS) ) {
				pass = asmpass;
			}

			cp = &passes[pass];
			if (pass != last_pass)
				(void) strcpy( outname, basename( inname )); /* Get its basename */
			else
				(void) strcpy( outname, nameptr->name);    /* get full original name */

			/* Add the temporary directory, and ensure suffix is correct */
			add_tmp( outname, cp->p_suffix, passes[last_pass].p_suffix);
			ret=command_line (cp->p_name, cp->p_optbuf, inname, outname);

			if ((ret != 0)
#ifdef QL
#ifdef QDOS
			&& !(ret == -1 && pass == GWASS_PASS) /* Patch for GWASS */
#endif /* QDOS */
#endif /* QL */
				){
#ifdef QL
#ifdef QDOS
				if (pass == GWASS_PASS) {
					char  errbuf[10];
					(void) itoa (ret, errbuf);
					printstr ("GWASS returned error code ",errbuf,"\n",NULL);
				}
#endif /* QDOS */
#endif /* QL */
				DBG(("MAIN",0x18,"command failed (%d) - tidying up",ret));
				(void) unlink (outname);   /* remove failed output file */
#ifdef QDOS
				if( no_continue()) {
					exit( ret );
				}
#endif /* QDOS */
				dolink = 0; /* A pass failed - forget linking */
				break;
			}
			/*
			 *	Delete any intermediate files
			 *	- but not _o's
			 */
			if( pass != start_pass ) {
				DBG(("MAIN",0x18,"Deleting intermediate file '%s'",inname));
				(void) unlink( inname );
			}
#ifdef QL
#ifdef QDOS
			if (pass == GWASS_PASS) {
				pass = COPY_PASS;
				(void) strcpy(strrpbrk (inname,"._"),"_o");
				if (strcmp(inname,outname)) {
					(void) command_line ("cp", "", inname, outname);
					(void) unlink (inname);
				}
				(void) strcpy(strrpbrk (inname,"._"),"_LST");
				DBG(("MAIN",0x18,"Deleting GWASS list file '%s'",inname));
				(void) unlink (inname);
				(void) strcpy(strrpbrk (inname,"._"),"_SYM");
				DBG(("MAIN",0x18,"Deleting GWASS symbol file '%s'",inname));
				(void) unlink (inname);
				pass = GWASS_PASS;
			}
#endif /* QDOS */
#endif /* QL */
			(void) strcpy( inname, outname); /* Output from previous pass goes into
									  next pass */
		}

		if( dolink && start_pass != LD_PASS ) {
			DBG(("MAIN",0x18,"adding file %s to list of link files",outname));
			add_option( LD_PASS, outname);
			add_option( DEL_PASS, outname);
		}
	}

	DBG(("MAIN",0x18,"Is link required?  dolink=%d",dolink));
	if(!dolink ) {
		exit (ret);
	}

	/*
	 *	We have now finished the compilation phase
	 *	Finally do the link if neccessary.
	 */
	DBG(("MAIN",0x14,"Start link phase"));
	if (! (ret = command_line (passes[LD_PASS].p_name,
							   passes[LDPRE_PASS].p_optbuf,
							   passes[LD_PASS].p_optbuf,
							   passes[LDPOST_PASS].p_optbuf))) {
		/*
		 *	delete any intermediate object 
		 *	files at this point
		 */
		for (p=passes[DEL_PASS].p_optbuf ; *p != 0 ; p += strlen(p) + 1 ) {
			while (*p == ' ') p++;	/* skip any leading spaces */
			if ( (p1 = strpbrk (p, " ")) != NULL) {
				*p1 = '\0'; 	/* Convert space to terminator */
			}
			(void) unlink (p);
		}
	} else {
		printstr("Link failed\n", NULL);
		exit(ret);
	}
	DBG(("MAIN",0x11,"CC finished - return code = %d",ret));
	exit (ret);
}
